/*
 * Jeffrey Friedl
 * Omron Corporation			オムロン（株）
 * Nagaokakyoshi, Japan			〒617長岡京下海印寺
 *
 * jfriedl@nff.ncl.omron.co.jp
 *
 * This work is placed under the terms of the GNU General Purpose License
 * (the "GNU Copyleft").
 */
#define PATCHLEVEL 7 /* patchlevel off the major version number */
const char version_string[] = "Version 1.08beta";
const char version_date  [] = "July 19, 1996";
const char compile_date  [] =  __DATE__;
const char author_name   [] = "Jeffrey Friedl";
const char contact_addr  [] = "jfriedl@omron.co.jp";
